#define MAX_ASYNC_DEPTH_PREINIT 15
#define MAX_ITER_NEWTON_PREINIT 15
#define BIAS_PREINIT 1e-2
#define BIAS_CUTE_PREINIT 1e-3
// biasCute = 0 --- always use cuteCalculation
// biasCute = 10 --- never

// when the size of the box is smaller than EPS_PREINIT then the box is declared undecided
// #define EPS_PREINIT 7.0 * 1e-3;
#define EPS_PREINIT 1e-5;  // the proof from our paper on planar cc with equal masses go through for 1e-5

#include <iostream>
#include <cstdio>
#include <list>
#include <time.h>
#include <map>
#include <string>
#include <chrono>


#include <capd/vectalg/Vector_inline.h>
#include <vector>
#include <future>
#include <thread>
#include <math.h>

#include "defo.h"
#include "cc-nb.h"
#include "solutions-tests.h"
#include "test-and-help-functions.h"
#include "estimates-functions.h"
#include "MyIntervalFunc.h"
#include "find-symmetry.h"
#include "searchalg.h"
#include "cuteCalc.h"

// File for binary output data
//ofstream outFileFailed;
// ofstream outFileTest;
// Files for Mathematica-format output data
ofstream mathFile;
ofstream mathListFile;

char zerosFile[30];       // the name for the file where all boxes coinating zeros (i.e. CC) found in the search stage are written - binary file
char undecidedFile[30];   // the name for the file where all undecided boxes in the search stage or the test stage are written - binary file
char zerosTestedFile[30]; // the name for the file where all boxes coinating zeros (i.e. CC) found in the testing stage are written - binary file
char foutfMath[30];       // the name of the file for some mathematica data - text file
char foutfMathList[30];   // the name of the file for some mathematica data - text file



std::atomic<int> cnt(0);
atomic<int> boxCount(0); // number of checked boxes

atomic<long> aprioriCounter(0); // tell how many times <checkAproriBounds failed
atomic<long> clusterCounter(0);
atomic<long> distanceCounter(0);
atomic<long> checkZeroCounter(0); // checkZero 0 \in f_i
atomic<long> KrawczykFailedCounter(0);
atomic<long> KrawczykZeroCounter(0);
atomic<long> KrawczykNoZeroCounter(0);
atomic<long> UEqICounter(0); //             checkUEqI  (I=U)
atomic<long> clusterImprovementCounter(0);
atomic<long> interQuadCounter(0);
atomic<long> methodFailedCounter(0);
atomic<long> methodZeroes(0);
atomic<long> isoBoxesCount(0);  // used in FINITENESS_TEST , to count how many times the isomorphism test worked


mutex updateCounter_mutex;
mutex krawczykSection_mutex;
mutex write_mutex;

bool equalMasses;
bool orderIncreasing = false;
// it makes sense only for equalMasses, N > 4
// orderIncreasing = true: x2 <= x3 <= ... <= x_{n-3} <= x_{n-1} it makes sense only for equalMasses, N > 4
// orderIncreasing = false: x2 >= ... >= x_{n-3} >= x_{n-1}
bool fixedOrderOfBodies;
// in equalMasses case this parameter has no meaning;
// otherwise this parameter tells whether to permute bodies (fixedOrderOfBodies = false) or not (fixedOrderOfBodies = true)
// if (!fixedOrderOfBodies) then init parameters are set by aprioriBoundsData, this part of entry parameters is ignored

ZeroSearchType methodUsed = ZeroSearchType::krawczyk;
NormalizationMethod normalizationMethod = NormalizationMethod::equalToZero; // y_{n-1}=0,  z_0=0, z_{n-1}=0
MyDouble eps = EPS_PREINIT;                                                 //  when in the subdivion algorithm we reeach the box  with max diam over active coordinates < eps , then we deem it undecide
MyDouble bias = BIAS_PREINIT;                                               // for Krawczyk method, if the size of box (max diam() coordinates) >= bias, then we do not call Krawczyk operator
MyDouble biasCute = BIAS_CUTE_PREINIT;                                      // for cuteCalculation: 0 -- always use cuteCalculation, ... 100 -- "never" use cuteCalculation
                                                                            // najlepszy eksperymentalnie dla N=4 to biasCute = 1e-4
                                                                            // dla N = 5 --- biasCute = 1e-3
                                                                            // dla N = 6 --- biasCute = 1e-4

int MAX_ITER_NEWTON = MAX_ITER_NEWTON_PREINIT;

int MAX_ASYNC_DEPTH = MAX_ASYNC_DEPTH_PREINIT;




// #include "testCuteCalc.cpp"  // some  tests of cuteCalculation function


#ifdef PLANAR
int initParameters(const char *inputFile,vector<MyDouble> &xInit, vector<MyDouble> &yInit,vector<MyDouble>  &massInit)
#endif
#ifdef SPATIAL
int initParameters(const char *inputFile,vector<MyDouble> &xInit, vector<MyDouble> &yInit, vector<MyDouble> & zInit, vector<MyDouble> & massInit)
#endif
//  Returns: number of bodies
// structure of the data file:
// * 1st line
// 0 -- different masses, full configuration space;
//      positions in the data file are ignored
// 1 -- equal masses, bodies positions within prescribed bounds, equalMasses=true
// 2 -- different masses, bodies positions within prescribed bounds , fixedOrderOfBodies=true
// * 2nd line
// number of bodies
// * next lines
//  different for planar and spatial case
//  SPATIAL
//  each line contains: intervals for x, y, z-coordinates of bodies and (in the case fo unequal masses) interval for masses
//  PLANAR
//   // * 3rd line
// intervals for x-coordinates of bodies
// * 4th line
// intervals for y-coordinates of bodies
// * 5th line
// intervals for masses of bodies
{
    int numBodies;
    
    char trash[240];

    ifstream plik(inputFile);
	if(!plik){
	  cout << "cannot open " << inputFile << endl;
	  exit(1);
	}
	
	
    xInit.clear();
    yInit.clear();
	#ifdef SPATIAL
    zInit.clear();
	#endif
    massInit.clear();
   
	
    if (plik.good() == true)
    {
        int massParam;
        plik >> massParam;
        plik.getline(trash, 240); // to eat endl - and the eventual comment following
		
		if((massParam<0) || (massParam>2))
		{
           cout << "Wrong input data -- mass parameter in the first line!" << endl;
           exit(3);
        }
		
        equalMasses = (massParam == 1);
        fixedOrderOfBodies = (massParam == 2);
		

        plik >> numBodies;
        plik.getline(trash, 240);
	
		
        if (equalMasses)
        {
		     MyInterval m= MyInterval(1.0) / numBodies;
            for (int i = 0; i < numBodies - 1; i++)
            {
                massInit.push_back(m.leftBound());
                massInit.push_back(m.rightBound());
            }
        }

        // used only in the case fo unequal masses
        MyInterval massSum = MyInterval(0.0);
        MyInterval massOfBodies[numBodies];
       
        //-------------------------
      
		
		#ifdef SPATIAL
		//  we read x,y,z-coordinates and mass
        for (int i = 0; i < numBodies - 1; i++)
        {
            if (plik.eof())
            {
                cout << "Wrong input data -- x!" << endl;
                exit(3);
            }
			MyDouble temp;
            plik >> temp;
            xInit.push_back(temp);
            plik >> temp;
            xInit.push_back(temp);
            plik >> temp;
            yInit.push_back(temp);
            plik >> temp;
            yInit.push_back(temp);
            plik >> temp;
            zInit.push_back(temp);
            plik >> temp;
            zInit.push_back(temp);
            if (!equalMasses)
            {
			    MyDouble temp1,temp2;
                plik >> temp1;
                plik >> temp2;
                massOfBodies[i] = MyInterval(temp1, temp2);
                massSum += MyInterval(temp1, temp2);
            }
            plik.getline(trash, 240);
        }
		#endif
		#ifdef PLANAR
		 //  we read x-coordinates
        for (int i = 0; i < numBodies - 1; i++)
        {
            if (plik.eof())
            {
                cout << "Wrong input data -- x!" << endl;
                exit(3);
            }
			MyDouble temp;
            plik >> temp;
            xInit.push_back(temp);
            plik >> temp;
            xInit.push_back(temp);
        }
        plik.getline(trash, 240);

         //  we read y-coordinates
        for (int i = 0; i < numBodies - 1; i++)
        {
            if (plik.eof())
            {
                cout << "Wrong input data -- y!" << endl;
                exit(3);
            }
			MyDouble temp;
            plik >> temp;
            yInit.push_back(temp);
            plik >> temp;
            yInit.push_back(temp);
        }
        plik.getline(trash, 240);
		 
		// we read masses 
		if(!equalMasses){
          
           
            for (int i = 0; i < numBodies - 1; i++)
            // the last mass is calculated as 1.0 - massSum
            {
                if (plik.eof())
                {
                    cout << "Wrong input data -- m!" << endl;
                    exit(3);
                }
				 MyDouble temp1, temp2;
                plik >> temp1;
                plik >> temp2;
                massOfBodies[i] = MyInterval(temp1, temp2);
                massSum += MyInterval(temp1, temp2);
            }
            plik.getline(trash, 240);
		}	
		#endif
	
        if (!equalMasses)
        {

            MyInterval lastMass = 1.0 - massSum;
            massOfBodies[numBodies - 1] = lastMass;

           

            // loading massOfBodies to massInit
            for (int j = 0; j < numBodies; j++)
            {
			    MyDouble temp;
                temp = massOfBodies[j].leftBound();
                massInit.push_back(temp);
                temp = massOfBodies[j].rightBound();
                massInit.push_back(temp);
            }
        }
       
        plik.close();
    }
    else
    {
        cout << "No access to the input data file!" << std::endl;
        exit(2);
    }

   

    return numBodies;
}



void writeReportCounters(ostream &fout)
// writes all counters to fout
{

    fout << endl;
    #ifndef FINITENESS_TEST
    fout << "The number of zeros in the method: " << methodZeroes << endl;
    #endif
	#ifdef FINITENESS_TEST
    fout << "The upper bound of zeros in the method: " << methodZeroes + isoBoxesCount << endl;
    fout << "The number of iso boxes: " << isoBoxesCount << endl;
	#endif
    if (methodFailedCounter > 0)
             fout << "The number of UNDECIDED cubes: " << methodFailedCounter << endl;
		    
    #ifdef FINITENESS_TEST
	 if (methodFailedCounter == 0)
	   fout << "FINITENESS PROVED" << endl;
	 else   
	   fout << "FINITENESS WAS NOT ESTABLISHED" << endl;
    #endif 		

            fout << endl;

    fout << "Tests usage:" << endl;
    fout << "checkAprioriBounds -- " << aprioriCounter << endl;
    fout << "clusterTest -- " << clusterCounter << endl;
    fout << "distanceTest -- " << distanceCounter << endl;
    fout << "checkZero -- " << checkZeroCounter << endl;
    fout << "krawczyk: methodFailed  -- " << KrawczykFailedCounter << endl;
    fout << "krawczyk: zeroIside -- " << KrawczykZeroCounter << endl;
    fout << "krawczyk: no zero inside -- " << KrawczykNoZeroCounter << endl;
    fout << "U = I -- " << UEqICounter << endl;
    fout << "improvement in clusterTest -- " << clusterImprovementCounter << endl;
    fout << "improvement in U = I Test -- " << interQuadCounter << endl;
    #ifdef  FINITENESS_TEST
	fout << "isomorphism test -- " << isoBoxesCount << endl;
    #endif        

            fout << endl
                 << "Number of checked boxes = " << boxCount << endl;
				 
	

   fout << endl << flush;
			
			
}

void ZeroAllCounters()
{
  cnt=0;
  boxCount=0; // number of checked boxes

  aprioriCounter=0; // tell how many times <checkAproriBounds failed
 clusterCounter=0;
 distanceCounter=0;
 checkZeroCounter=0; // checkZero 0 \in f_i
 KrawczykFailedCounter=0;
 KrawczykZeroCounter=0;
 KrawczykNoZeroCounter=0;
 UEqICounter=0; //             checkUEqI  (I=U)
 clusterImprovementCounter=0;
 interQuadCounter=0;
 methodFailedCounter=0;
 methodZeroes=0;
 isoBoxesCount=0;
}

void writeGlobalSettings(ostream &fout,int NUM_BODIES)
{

    #ifdef FINITENESS_TEST
	fout << "FINITENESS TEST " << endl;
	#endif

    fout << "Number of bodies = " << NUM_BODIES << endl;
    if (equalMasses)
        fout << "EQUAL MASSES" << endl;
    else if (fixedOrderOfBodies)
        fout << "UNEQUAL MASSES and FIXED ORDER OF BODIES!" << endl;
    else
        fout << "UNEQUAL MASSES --- we look for ALL CC "
             << " (computations with permutations of bodies)!" << endl;
    fout << endl;
    fout << "Accuracy eps = " << eps;
    fout << ", bias = " << bias;
    fout << ", biasCute = " << biasCute << endl;
    fout << "MAX_ASYNCH_DEPTH = " << MAX_ASYNC_DEPTH << endl;
    fout << "----------------------------------------------" << endl;
    fout << endl
         << endl
         << flush;
		 
}


void writeUndecidedBoxes(int NumBodies,const char* undecidedBoxesFile)
// the function gives report on undecided boxes
//  and also writes in Mathematica format to files 'mathFile','mathListFile' which are globally defined ostream's
{
  vector<Bodies> undecidedBoxes;
  int ccNo;
  if ((ccNo = readBodiesFromBinaryFile(NumBodies,undecidedBoxesFile, undecidedBoxes)) > 0)
  {
    cout << "the number of undecided boxes from the binary file " << ccNo << endl   << flush;
    mathListFile << "------- undecided ----------";
    for (int k = 0; k < ccNo; k++)
    {
        undecidedBoxes[k].writeCCformatMath(mathFile,mathListFile);
        cout << "---------------------" << endl;
        undecidedBoxes[k].printData(true); // writes distances from 0
        cout << flush;
    }
  }
}

 bool basicSearchAndTest(Bodies bodies, bool doSearch, bool doTest,const char*  zerosFile,
     const char*  undecidedFile,const char*  zerosTestedFile)
 //  zerosFile - the name for the file where all boxes coinating zeros (i.e. CC) found in the search stage are written - binary file
 //            it is an output from the search stage and serves as the input to the testing stage             
//   undecidedFile - the name for the file where all undecided boxes in the search stage or the test stage are written - binary file
//            it is an output from the search stage and also an output from the test stage
// zerosTestedFile - the name for the file where all boxes coinating zeros (i.e. CC) found in the testing stage are written - binary file
//                   this is rather an auxlilary file , can be used for diagnostic purposes
//            
 // uses global variables: all counters
 {
   int numBodies=bodies.N;
   #ifdef PLANAR
   MyDouble midMass;
   for (int i = 0; i < numBodies; i++)
   {
       midMass = (bodies.mass(i).rightBound() + bodies.mass(i).leftBound()) / 2.0;
       mathFile << midMass << "  ";
       mathListFile << midMass << ",  ";
   }
   mathFile << endl   << endl;
   mathListFile << endl   << endl;
   #endif
		
		
   

   if (doSearch)
   {
   
      cout << "Input data:" << endl;
      bodies.printData();
      cout << endl; 
    
       ofstream outFileSearch;
	   ofstream outFileFailed;  
       outFileSearch.open(zerosFile, ios::binary | ios::out | ios::trunc);
       outFileFailed.open(undecidedFile, ios::binary | ios::out | ios::trunc);

       try
       {
          searchCC(bodies, doSearch, outFileSearch, outFileFailed);
                //  solutions (boxes) found are written to   <outFileSearch>, the undecided boxes to <outFileFailed>
                // in binary format
       }
       catch (...)
       {
            cout << "Exception!" << endl;
			exit(1);
       }
       outFileFailed.close();
       outFileSearch.close();
       writeReportCounters(cout);

 
   }
      

    
    if (doTest)
    {
     
       int uniqueSol = 0; // the number of different solutions 

       int ccNo;
       vector<Bodies> ccSolutions;
       vector<int> markedSolutions;

       if ((ccNo = readBodiesFromBinaryFile(numBodies, zerosFile, ccSolutions)) > 0)
       {
          identifyDifferentSolutions(equalMasses, ccNo, ccSolutions, markedSolutions, uniqueSol,zerosTestedFile,undecidedFile);   // different solutions are returned
				                                 // as those indices i, such that markedSolutions[i]==0
          if (uniqueSol < 1)  // all solutions do not satisfy our order conditions
          {
            cout << endl;
            cout << "Ups!!! distanceTest failed!!!" << endl;
			// we report these solution for possible inspection 
            for (int k = 0; k < ccNo; k++)
            {
              ccSolutions[k].writeCCformatMath(mathFile, mathListFile);
              cout << "---------------------" << endl;
              ccSolutions[k].printData(true); // writes distances from 0
              cout << flush;
            }
          }
          writeDiffCCandTestSymmetry(ccNo, ccSolutions, markedSolutions);  // writes to output file different CC with the full information
				                                                 // about their symmetries
       }
       else
       {
         uniqueSol = 0;
         cout << "No CC found!" << endl;
       }
	   
	    if (methodFailedCounter > 0)
		{
           writeUndecidedBoxes(numBodies,undecidedFile);    
        }
	   
    }

    cout << "-------------------------------------------------" << endl  << endl;
	return true;
}


void nextPermOfIndices(int perm[], int n, int point)  
// on output: perm[point] = n-1; perm[n-2] = point; perm[n-1] = n-2
//    and perm[i]=i, for i\neq point,n-2,n-1
//   we realize the following cycle point ->  n-1 ->  n-2 -> point
{
    for (int i = 0; i < n - 1; i++)
        perm[i] = i;
    perm[n - 1] = n - 2;
    perm[n - 2] = point;
    perm[point] = n - 1;
}


void searchAndTestWithPermutations(Bodies baseBodies,bool searchStage,bool testStage,const char*  zerosFile,
     const char*  undecidedFile,const char*  zerosTestedFile)
// it makes sense only if !equalMasses && !fixedOrderOfBodies
{
  int numBodies=baseBodies.N;
  
  Bodies bodies(baseBodies);
  
  int perm[ numBodies];
  for (int i = 0; i <  numBodies; i++)
        perm[i] = i;
		
  for (int point = 0; point <  numBodies; point++) // this loop is indexed with the body which is the furthers from the origin
                                                     // it is executed many times in the case of not equal masses and not fixed permutation
													 // MYSLE ZE TAK JAK MAMY w distanceTest - to to sa niewystarczajace permutacje, bo trzeba
													 // jeszcze wybierac cialo q0
  {
     
     cout << "Current permutation: ";
     mathFile << "Current permutation of masses: " << endl;
     mathListFile << "Current permutation of masses: " << endl;
     for (int i = 0; i < numBodies; i++)
     {
       cout << perm[i] << "  ";
       mathFile << perm[i] << "   ";
       mathListFile << perm[i] << "   ";
     }
     cout << endl;
     mathFile << endl;
     mathListFile << endl;
      
    // ZeroAllCounters();  // we reset all counters

     basicSearchAndTest(bodies,searchStage,testStage,zerosFile,undecidedFile,zerosTestedFile);
       
       

      
     nextPermOfIndices(perm,numBodies, point);
            // on output: perm[point] = n-1; perm[n-2] = point; perm[n-1] = n-2
            // it is desirable that the last body ((n-1)-th) has to be the heaviest among q(0)... q(N-3), but at the moment we do not force this
      bodies = baseBodies;
      permuteMassesOfBodies(bodies, perm);
      setAprioriBoundsData(bodies);
   }   
}


int main(int argc, char *argv[])
{
    //superTestCute();
    //exit(1);

    initIntervalGlobals(); // initiates sqrt2, sqrt3, pi - these variables have to be initiated somewhere in main,
                      // trying to initiate them as global variables may cause "segmentation fault" on some systems

    if (argc < 4)
    {
        cerr << "Wrong number of parameters!" << endl;
        exit(1);
    }

    char idi = *argv[2]; // instance id
    if (idi < '0' || idi > '9')
    {
        cerr << "Wrong instance id!" << endl;
        exit(1);
    }

    // searchStage = true -- searching mode, false -- testing mode
    // in the searching mode we write found solutions into a binary file;
    // in testing mode only  to 'cout'; in order to get the result in a text file we should redirect the output
    bool searchStage;
    bool testStage;
    // set searchStage = true, if the parameter 's' (test) is given.
    // set testStage = true, if the parameter 't' (test) is given.
    char *par = argv[1];
    searchStage = par[0] == 's';
    testStage = par[0] == 't' || par[1] == 't';

    if (argc == 5)
        MAX_ASYNC_DEPTH = atoi(argv[4]);

    char inputFile[30]; // the name for the basic input data file - text file

    sprintf(inputFile, "data/init-data-%s-%s.txt", argv[2], argv[3]);
    sprintf(zerosFile, "data/cc-zeros-%s-%s.dat", argv[2], argv[3]);
    sprintf(zerosTestedFile, "data/cc-tested-zeros-%s-%s.dat", argv[2], argv[3]);
    sprintf(undecidedFile, "data/cc-undecided-%s-%s.dat", argv[2], argv[3]);
    sprintf(foutfMath, "data/math-%s-%s.txt", argv[2], argv[3]);
    sprintf(foutfMathList, "data/math-%s-%s-list.txt", argv[2], argv[3]);

    mathFile.open(foutfMath, ios::out | ios::trunc);
    mathFile.setf(ios::fixed, ios::floatfield);
    mathFile.precision(10);

    mathListFile.open(foutfMathList, ios::out | ios::trunc);
    mathListFile.setf(ios::fixed, ios::floatfield);
    mathListFile.precision(10);

    vector<MyDouble> xInit, yInit, zInit, massInit;

#ifdef PLANAR
    int NUM_BODIES = initParameters(inputFile, xInit, yInit, massInit);
#endif
#ifdef SPATIAL
    int NUM_BODIES = initParameters(inputFile, xInit, yInit, zInit, massInit);
#endif
    
    if (atoi(argv[2]) != NUM_BODIES)
    {
        cout << "Wrong input data! the numbers of bodies are incorrect" << endl;
        exit(2);
    }

#ifdef PLANAR
    Bodies bodies(NUM_BODIES, true, xInit, yInit, massInit);
#endif
#ifdef SPATIAL
    Bodies bodies(NUM_BODIES, true, xInit, yInit, zInit, massInit);
#endif


    if (!equalMasses)
        if (!fixedOrderOfBodies)
            setAprioriBoundsData(bodies);

    if (bias < eps)
        bias = eps;

    #ifdef FINITENESS_TEST
	testStage=false;   // we just compute the boxes wherewhere we have at most one solution, hence  the testing does not make sense
	if(testStage){
	  cout << "in the FINITENESS_TEST mode the testing does not make sense" << endl << flush;
	  exit(1);
	}  
	#endif


    std::cout.precision(10);
    writeGlobalSettings(cout,NUM_BODIES);
	
	
	if(equalMasses  || fixedOrderOfBodies)
	{
	    basicSearchAndTest(bodies,searchStage,testStage,zerosFile,undecidedFile,zerosTestedFile);
	}else
	{
       searchAndTestWithPermutations(bodies,searchStage,testStage,zerosFile,undecidedFile,zerosTestedFile);
    }
    mathFile.close();
    mathListFile.close();
    return 0;
}
