#include "defo.h"
#include "krawczyk-newton.h"
#include "MyIntervalFunc.h"
#include "bodies.h"


#include "cc-nb.h"




using namespace std;
using namespace capd;

/* ================= */
/*  Newton's method  */
/* ================= */

MyVector NewtonOperator(const MyMatrix &invDF,const Bodies &bodies) 
//  returns: x0 - invDF * Fx0
{
  int dim=bodies.dimRedConfSpace();
  
  MyVector Fx0(dim);
  MyVector x0(dim);
  Bodies *midBodies = bodies.calculateMidBodies();
  x0=midBodies->toReducedSpace() ;
  midBodies->compFRed(Fx0);
  delete midBodies;
  
  return x0 - invDF * Fx0;
}

ResultType newtonMethod(Bodies copyBodies, Bodies &returnValues, bool &intersec,  MyMatrix &DFF)
{
 
  return ResultType::methodFailed;
}

/* =================== */
/*  Krawczyk's method  */
/* =================== */



MyVector krawczykOperator(const MyMatrix &C,const MyMatrix &DFF, const Bodies &bodies, MyVector &chZ)
// Input:
//   C - some linear isomorphism
//   DFF - D Freduced with respect to coordinates parameterizing the reduced space
//   bodies - the set on which the Krawczyk operator is computed   
// on output:
//   returns  x0 - C * F(x0) + (I - C*DFF(x)) * (x-x0),
//       where x = bodies in reduced coordinates and x0-is mid point of x 
 // chZ --- mean Value evaluation of freduced(bodies), 
{

  int  dimRCS=bodies.dimRedConfSpace();

  MyVector Fx0(dimRCS);
  MyVector x0(dimRCS),x(dimRCS);
  MyVector xMinusx0(dimRCS);
 
  Bodies *midBodies = bodies.calculateMidBodies();
  x0=midBodies->toReducedSpace();
  midBodies->compFRed(Fx0);
  delete midBodies;
 
  MyMatrix CDFF = -C * DFF;
  for (int i = 0; i < dimRCS; i++)
    CDFF[i][i] += 1.0;
  // now CDFF = I-C*DFF

  x=bodies.toReducedSpace();
  xMinusx0=x-x0;

  chZ = Fx0 + DFF * xMinusx0; // Mean Value Form

  return x0 - C * Fx0 + CDFF * xMinusx0;
}



//*/
//-----------------------------------------------------------------------------------
ResultType krawczykMethod(Bodies boxBodies, Bodies &returnValues, bool &intersec,  MyMatrix &DFF)
{
   // the Krawczyk method is applied to <boxBodies> 
   // On input: DFF = DfRed(boxbodies) - it is supplied because this is a costly function and is used elsewhere
  // On output:
  //    intersec = true  when the method fails but still KrawczykMethod succeded in reducing the size of the search box
  //               false otherwise
  //   returnValues - is the result of Newton/Krawczyk operator interations (Nx \cap x if the intersection happens to be non-empty) 
  //    DFF contains DfRed(returnValues) if the flag forcing the recomputation of Df was turned on





 // ResultType retVal;
  int dimRCS=boxBodies.dimRedConfSpace();
  MyMatrix C(dimRCS, dimRCS); 
  MyVector Nx(dimRCS), Y(dimRCS), chZ(dimRCS);
  MyDouble accuracy = 1e-6; // proceed until the interval is sufficiently small
 
  MyDouble volboxBodies;

  bool foundSolution = false;
  intersec = false;

  #ifdef TESTY
  cout << "po wejsciu do metody Krawczyka " << endl;
  cout << "dimRCS=" << dimRCS << endl;
  MyVector xboxb(dimRCS),Fxboxb(dimRCS);
  xboxb=boxBodies.toReducedSpace();
  boxBodies.compFRed(Fxboxb);
  cout <<  "xboxb=" << xboxb << endl;
  cout << "FRed(xboxb)=" << Fxboxb<< endl;
  cout << endl;
  #endif



  //  Computation of matrix C
  //computeDFRedMatrix(boxBodies, DFF);
  DMatrix midDF(dimRCS, dimRCS);
  midDF = capd::vectalg::convertObject<DMatrix>(DFF);
  try
  {
    DMatrix doubleC(dimRCS, dimRCS);
    doubleC = capd::matrixAlgorithms::gaussInverseMatrix(midDF);
    C = capd::vectalg::convertObject<MyMatrix>(doubleC);
	if(!isomorphismTest(C))
      return ResultType::methodFailed;
  } 
  catch (std::runtime_error)
  {
    return ResultType::methodFailed;
  }
  // now C\appox mid(DFF)^{-1}
  
  
  bool finished=false;
  int loopC = 0;
  while(!finished) 
  {
    Nx = krawczykOperator(C, DFF, boxBodies, chZ);
	
	#ifdef TESTY
	cout << "Result of KrawczykOperator:" << endl;
    cout << Nx << endl;
	cout << endl;
	#endif
	
	
    for (int i = 0; i < dimRCS - 1; i++)   // we check whether the mean value evaluation performed in <krawczykOperator> shows that there is no zero in <boxBodies>
    {
      if (!isSingular(chZ[i]))
	  {
	    finished=true;
		// retVal=ResultType::noZeroInSet;
        return ResultType::noZeroInSet; // there is no solution
	  }	
    }
	
	
    if(!finished){
      if (cubeSubSetInterior(Nx, boxBodies))
      {
        foundSolution = true;
      }
      else if (cubeEmptyIntersection(Nx, boxBodies))
	  {
	    finished=true;
		// retVal=ResultType::noZeroInSet;
        return ResultType::noZeroInSet; // there is no solution
      }
	}  
	if(!finished)
	{
      if (cubeSuperSet(Nx, boxBodies))  // bodies \subset Nx
      {
	    finished=true;
        returnValues = boxBodies;
		// retval=foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
        return foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
       
      }
    }
	if(!finished)
	{
	  intersec = true;
	
      volboxBodies = volume(boxBodies);
	  
      Y = cubeIntersection(Nx, boxBodies);
      updateBodies(Y, boxBodies);  // boxBodies=Y (in the reduced configuration space)
     
	  // now we decide if it is worth to continue with the iteration
      MyDouble v = volume(boxBodies);
      if (1.0 - v / volboxBodies < 0.3 || power(v, 1.0 / dimRCS) <= accuracy)
      {
	    finished=true;
        returnValues = boxBodies;
        return foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
      }

      // dla n = 4 przeliczanie DFF po przecieciu w krawczyku jest zbyt kosztowne!!!
      #ifdef RECOMPUTE_DFF_krawczyk 
      computeDFRedMatrix(boxBodies, DFF); 
      #endif
	}  
    loopC++;
    finished=finished || (loopC >= MAX_ITER_NEWTON);
  } // end while (!finished)

  returnValues = boxBodies;
  return foundSolution ? ResultType::uniqueZero : ResultType::methodFailed;
}




