#ifndef krawnet_h
#define krawnet_h

#include "defo.h"
#include "bodies.h"




enum class ResultType
{
    uniqueZero=1,     // there is exactly one zero
    methodFailed=2,   // the matrix is singular or Krawczyk/Newton operator result contains the initial condition
    noZeroInSet=3,    // no zero found
    collisionFound=4, // collision found
    twoZeros=5,       // two zeros found
    zeroFound=6       // zero found, but already processed
};


enum class ZeroSearchType
{
    krawczyk,
    newton
};


ResultType newtonMethod(Bodies copyBodies, Bodies &returnValues, bool &intersec,  MyMatrix &DFF);


ResultType krawczykMethod(Bodies boxBodies, Bodies &returnValues, bool &intersec,  MyMatrix &DFF);
 // the Krawczyk method is applied to <boxBodies> 
   // On input: DFF = DfRed(boxbodies) - it is supplied because this is a costly function and is used elsewhere
  // On output:
  //    intersec = true  when the method fails but still KrawczykMethod succeded in reducing the size of the search box
  //               false otherwise
  //   returnValues - is the result of Newton/Krawczyk operator interations (Nx \cap x if the intersection happens to be non-empty) 
  //    DFF contains DfRed(returnValues) if the flag forcing the recomputation of Df was turned on



#endif
